@@ -15,24 +15,30 @@
 """Flask module."""
 
 from functools import wraps
-from typing import Optional, Union, List
-from flask import current_app, Flask, request
+from typing import List, Optional, Union
+from urllib.parse import urlparse
+
+from flask import Flask, current_app, request
 from flask.helpers import make_response
 from flask.wrappers import Response
-from urllib.parse import urlparse
 from werkzeug.exceptions import HTTPException
 
-from .config import Config
+from .cache import Cache
 from .client import DefaultClient, NewDefaultClient
-from .errors import Error as IAMError, ClientTokenGrantError, GetClientInformationError, StartLocalValidationError, \
-    TokenRevokedError, UserRevokedError, ValidateAndParseClaimsError, ValidatePermissionError
-from .http_errors import InsufficientPermissions, InternalServerError, InvalidRefererHeader, UnauthorizedAccess, \
-    SubdomainMismatch
+from .config import Config
+from .errors import ClientTokenGrantError
+from .errors import Error as IAMError
+from .errors import (GetClientInformationError, StartLocalValidationError,
+                     TokenRevokedError, UserRevokedError,
+                     ValidateAndParseClaimsError, ValidatePermissionError)
+from .http_errors import (InsufficientPermissions, InternalServerError,
+                          InvalidRefererHeader, SubdomainMismatch,
+                          UnauthorizedAccess)
 from .models import JWTClaims, Permission
 
-
 # ---------- Exceptions ---------- #
 
+
 class HTTPError(HTTPException):
     def __init__(self, http_code: int, error_code: int, message: str, description: Optional[str] = None) -> None:
         super().__init__(description)
@@ -93,6 +99,7 @@ class IAM:
     """
 
     def __init__(self, app: Union[Flask, None] = None) -> None:
+        self.client_info_cache = Cache(ttl=60)
         self.app = app
         if app is not None:
             self.init_app(app)
@@ -214,7 +221,33 @@ class IAM:
             bool: Is referer header valid or not
         """
         try:
-            client_info = self.client.GetClientInformation(jwt_claims.Namespace, jwt_claims.ClientId)
+            # Cache implementation to handle race conditions during IAM URL changes
+            # When IAM URL is updated, there might be existing valid JWTs that were
+            # issued with the old URL. This cache ensures those tokens can still be
+            # validated during the transition period without making redundant requests
+            # to IAM for the same client information.
+                    
+            # Create cache key using namespace and client ID from JWT claims
+            # This combination uniquely identifies the client across IAM URL changes
+            cache_key = f"{jwt_claims.Namespace}:{jwt_claims.ClientId}"
+                    
+            # Try to get client info from cache first to avoid unnecessary IAM requests
+            # during the URL transition period. This is particularly important when
+            # handling multiple requests with JWTs issued under the old URL.
+            client_info = self.client_info_cache.get(cache_key)
+
+            if client_info is None:
+                # Cache miss - need to fetch from IAM
+                # This will use the current IAM URL configuration, but the response
+                # will be cached to handle subsequent requests that might still be
+                # using JWTs issued with the old URL
+                client_info = self.client.GetClientInformation(jwt_claims.Namespace, jwt_claims.ClientId)
+                if client_info:
+                    # Store successful response in cache
+                    # This ensures we can handle subsequent requests with old JWTs
+                    # without making additional IAM requests during the URL transition
+                    self.client_info_cache[cache_key] = client_info
+
         except GetClientInformationError:
             return False
 
